# 2016 

############################################################################################################
#
#  db_connector_runner - control machinery that manages the instantiation of a DB Connector
#                        and triggers the operation that retrieves data from Agile Central
#                        and deposits the data in a database or group of CSV files.
#
############################################################################################################

import sys, os
import time

from dbeif.utils.klog       import ActivityLogger
from dbeif.utils.konfabulus import Konfabulator
from dbeif.connector        import DBConnector

from dbeif.utils.eif_exception import ConfigurationError, NonFatalConfigurationError
from dbeif.utils.eif_exception import OperationalError, logAllExceptions
from dbeif.utils.metrixeur     import metrixeur

############################################################################################################

ARCHITECTURE_ACRONYM = 'eif'
__version__ = "1.1.0"

EXISTENCE_PROCLAMATION = """
************************************************************************************************************************

     CA Agile Central DatabaseConnector  version %s  starting at: %s
           curdir: %s
          command: %s

************************************************************************************************************************
"""

STD_TS_FMT = '%Y-%m-%d %H:%M:%S Z'

############################################################################################################

class DBConnectorRunner():
    def __init__(self, args):
        """
            An instance of this class is responsible for basic command line parsing,
            pulling off the name of the target config file.
            Once that item is taken care of, the instance obtains an instance of a
            Konfabulator, a AuxiliaryClassLoader instance and and instance of an DBConnector.
            The Konfabulator is delegated the task of reading the config file.
            These instances are then provided to the DBConnector instance.
            This instance then directs the DBConnector to run which involves obtaining
            a spoke instance to "source" the data from AgileCentral and a spoke instance
            to deposit the data in a csv and/or target database.
        """
        options = [item for item in args if item.startswith('-')]
        args    = [item for item in args if item not in options]
        if len(args) < 2:
            problem = ("Insufficient command line args, must be at least a config file name.")
            raise ConfigurationError(problem)
        self.program = args.pop(0)

        self.default_log_file_name = True
        self.cleartext_flag = False

        # this only allows a logfile spec of --logfile=<some_file_name>
        # TODO: consider allowing additional syntax  logfile=foo.log or --logfile foo.log
        log_spec = [opt for opt in options if opt.startswith('--logfile=')]
        if log_spec:
            self.logfile_name = log_spec[0].replace('--logfile=', '')
            self.default_log_file_name = False
            #for spec in log_spec:
            #    args = [arg for arg in args if arg != spec]
        self.encode_credentials = True
        clear_text_creds = [opt for opt in options if opt in ['-c', '--c'] or opt.startswith('--cleartext')]
        if clear_text_creds:
            self.cleartext_flag = True

        self.config_file_name = args.pop(0)
        if self.default_log_file_name:  # set it to first config file minus any '_config.yml' portion
            self.logfile_name = "log/%s.log" % self.config_file_name.replace('.yml', '').replace('_config', '')
            try:
                if not os.path.exists('log'):
                    os.mkdir('log')
            except Exception as msg:
                sys.stderr.write("Unable to locate or create the log sub-directory, %s\n" % msg)
                raise Exception

        self.log = ActivityLogger(self.logfile_name)
        logAllExceptions(True, self.log)
        self.preview = False
        self.connector = None
        self.extension = {}

    def proclaim_existence(self):
        now = time.strftime(STD_TS_FMT, time.gmtime(time.time()))
        proclamation = EXISTENCE_PROCLAMATION % (__version__, now, os.getcwd(), self.program)
        self.log.write(proclamation)


    def run(self):
        self.proclaim_existence()

        try:
            config_file_path = 'configs/%s' % self.config_file_name
            # if not config_file_path:
            #     raise ConfigurationError("No config file for '%s' found in the config subdir" % self.config_file_name)
            lf_name = "log/%s.log" % self.config_file_name.replace('.yml', '').replace('_config', '')
            self.log = ActivityLogger(lf_name)
            logAllExceptions(True, self.log)
            self._operateService(config_file_path)
        except Exception as msg:
            self.log.error(msg)
        self.log.info('run completed')


    def _operateService(self, config_file_path):
        config_name = config_file_path.replace('configs/', '')
        started = finished = elapsed = None
        self.connector = None
        self.log.info("processing to commence using content from %s" % config_file_path)

        last_conf_mod = time.strftime(STD_TS_FMT, time.gmtime(os.path.getmtime(config_file_path)))
        conf_file_size = os.path.getsize(config_file_path)
        self.log.info("%s last modified %s,  size: %d chars" % (config_file_path, last_conf_mod, conf_file_size))
        config = self.getConfiguration(config_name)
        if self.cleartext_flag:
            self.log.info("Running connector with cleartext flag. SecurityLevel will not be applied to credentials in cleartext")
        config.applySecurityPolicy()


        this_run = time.time()  # be optimistic that the reflectBuildsInAgileCentral service will succeed
        self.connector = DBConnector(config, self.log)
        self.log.debug("Got a DBConnector instance, calling the DBConnector.run ...")
        status, statistics = self.connector.run()
        if not statistics:
            self.log.info('No qualifying Agile Central data was retrieved, check your config file for the selection criteria.')
            return

        finished = time.time()
        elapsed = int(round(finished - this_run))
        self.logServiceStatistics(config_name, statistics, elapsed)

        if self.preview:
            self.log.info("Preview mode in effect, no data transferred")
            return

        # get the goods from the metrixeur and log the interesting stuff
        tracked_events = metrixeur.spill()
        for item in tracked_events:
            self.log.debug(item)


    def logServiceStatistics(self, config_name, statistics, elapsed):
        """
            what we intend to append to the log...
               foo_config.yml: dddd records processed for <AgileCentral entity name>
            and a line with the elapsed time taken in human readable form  
            (elapsed is in seconds (a float value))
        """
        preview_reminder = "(Preview Mode)" if self.preview else ""
        #for stat_item in statistics:
        for entity_name, wp_stats in statistics.items():
            for wp_dimension, stat in wp_stats.items():
                wksp, proj = wp_dimension
                factoid = "%3d %s items processed with %d fields for %s - %s" % (stat['Number of Records'], entity_name, stat['Fields Per Record'], wksp, proj)
                self.log.info("%s: %s %s" % (config_name, factoid, preview_reminder))
        hours, rem = divmod(elapsed, 3600)
        mins, secs = divmod(rem, 60)
        if hours > 0:
            duration = "%s hours %s minutes %s seconds" % (hours, mins, secs)
        else:
            if mins > 0:
                duration = "%s minutes %s seconds" % (mins, secs)
            else:
                duration = "%s seconds" % secs
        self.log.info("%s: service run took %s" % (config_name, duration))

    def getConfiguration(self, config_file):
        config = None
        try:
            config = Konfabulator(config_file, self.log, self.cleartext_flag)
        except NonFatalConfigurationError as msg:
            pass  # info for this will have already been logged or blurted
        except Exception as msg:
            raise ConfigurationError("Syntax error found in your configuration file:\n %s" % msg)
        svc_conf = config.topLevel('Service')
        self.preview = False
        if svc_conf and svc_conf.get('Preview', None) == True:
            self.preview = True
        self.log_level = 'Info'
        if svc_conf:
            ll = svc_conf.get('LogLevel', 'Info').title()
            if ll in ['Fatal', 'Error', 'Warn', 'Info', 'Debug']:
                self.log_level = ll
                self.log.setLevel(self.log_level)
            else:
                pass  # bad LogLevel specified

        return config
