import os,sys
from datetime import datetime
import pyodbc

from dbeif.connection import Connection
from dbeif.utils.eif_exception import OperationalError, ConfigurationError

__version__ = '1.1.0'

class MSSQLDBConnection(Connection):
    def __init__(self, config, logger, metrics_recorder=None):
        self.config = config
        self.mr     = metrics_recorder
        super().__init__(logger)
        self.backend_version = None
        self.internalizeConfig(config)

    def name(self):
        return 'dbeif.DBConnection'

    def version(self):
        return __version__


    def internalizeConfig(self, config):
        super().internalizeConfig(config)

        self.server       = config.get('Server', False)
        self.port         = config.get('Port',  80)
        self.type         = config.get('Type',   None)
        #self.user         = config.get('User',   None)
        #self.password     = config.get('Password', None)
        self.driver       = config.get('Driver', None)
        self.trusted_connection = config.get('Trusted_Connection', 'yes')
        self.db_name      = config.get('Name',   None)
        self.populate     = config.get('Populate', True)
        self.save_csv     = config.get('SaveCSV',  False)
        self.drop_tables  = config.get('DropExistingTables', False)
        self.stamp_tables = config.get('TimeStampTables', False)
        self.debug        = config.get("Debug",  False)

        # self.proxy = None
        # if self.proxy_server:
        #     self.proxy  = "%s://%s:%s" % (self.proxy_protocol, self.proxy_server, self.proxy_port)
        #     if self.proxy_username and self.proxy_password:
        #         self.proxy  = "%s://%s:%s@%s:%s" % (self.proxy_protocol, self.proxy_username, self.proxy_password, self.proxy_server, self.proxy_port)

        valid_config_items = ['Type','Name','Server', 'Port', 'User', 'Password','Populate',
                              'DropExistingTables','TimeStampTables','SaveCSV','Debug',
                              'Driver','Trusted_Connection']
        invalid_config_items = [item for item in config.keys() if item not in valid_config_items]
        if invalid_config_items:
            problem = "DBConnection section of the config contained these invalid entries: %s" % ", ".join(invalid_config_items)
            raise ConfigurationError(problem)

    def connect(self):
        try:
            conn_str = "DRIVER=%s;SERVER=%s;DATABASE=%s;Trusted_Connection=%s;" \
                       %(self.driver, self.server, self.db_name, self.trusted_connection)
            db = pyodbc.connect(conn_str)
            self.db = db
            self.cursor = self.db.cursor()
        except (pyodbc.Error, Exception) as ex:
            mssql_error = ex.args[1]
            problem = "Unable to connect to MSSQL at %s,  %s. Check driver name, credentials and database name" % (self.server,mssql_error)
            raise ConfigurationError(problem)

    def disconnect(self):
        self.cursor.close()
        self.db.close()

    def getBackendVersion(self):
        self.cursor.execute('SELECT @@version')
        version_str = self.cursor.fetchone()
        self.backend_version = version_str
        return self.backend_version

    def validate(self, entities):

        timestamp = datetime.now().isoformat().replace(':', '-').replace('-', '_').split('.')[0]
        self.table_stamp = timestamp if self.stamp_tables else None

        # are there tables present in the db that match the set to be processed?
        # if yes, is DropExistingTables False? then fail validate

        try:
            #self.cursor.execute("SELECT * FROM sys.tables")
            sql = "SELECT * FROM %s.INFORMATION_SCHEMA.TABLES WHERE TABLE_TYPE = 'BASE TABLE'" % self.db_name
            self.cursor.execute(sql)
            table_names = [table[2] for table in self.cursor.fetchall()]
        except (pyodbc.Error, Exception) as e:
            msg = e.args[1]
            print(msg)
            raise ConfigurationError("Unable to successfully query your MSSQL DB for table names in the %s database" % self.db_name)


        self.existing_tables = [self.sanitizeTableName(entity) for entity in entities
                                if self.sanitizeTableName(entity) in table_names]

        if self.table_stamp:
            return True

        if self.existing_tables:
            if self.drop_tables:
                self.log.warning('You have tables existing in the DB that will be dropped: %s' % ", ".join(self.existing_tables))
            else:
                self.log.error("Tables already exist, cannot proceed.  These tables need to be dropped or renamed: %s" % (", ".join(self.existing_tables)))
                self.log.error("Refer to the documentation about the TimeStampTables and DropExistingTables configuration optional settings.")
                return False

        return True


    def prepareDatabase(self, table_name):
        # if DropExistingTables True, drop the tables for the configured set of entities
        # use the self.existing_tables
        sanitized_table_name = self.sanitizeTableName(table_name)
        sanitized_table_name = "%s_%s" % (sanitized_table_name, self.table_stamp) if self.table_stamp else sanitized_table_name
        try:
            if self.drop_tables and sanitized_table_name in self.existing_tables:
                self.log.debug("Dropping table: %s" % sanitized_table_name)
                sql = 'DROP TABLE %s' %(sanitized_table_name)
                self.cursor.execute(sql)
        except Exception as e:
            msg = e.args[1]
            raise OperationalError("Unable to DROP TABLE %s, %s" % (sanitized_table_name, msg))

    # User is reserved word in MSSQLServer
    def sanitizeTableName(self, table_name):
         return 'Users' if table_name == 'User' else table_name.replace(" ", "")

    def sanitizeColumnName(self, column_name):
         return 'UserID' if column_name == 'User' else column_name

    def createTableAndColumns(self, table, attributes):
        temp_column = "tempcolumndeletelater"
        sanitized_table_name = self.sanitizeTableName(table)
        sanitized_table_name = "%s_%s" % (sanitized_table_name, self.table_stamp) if self.table_stamp else sanitized_table_name

        try:
            self.log.debug("Creating table: %s" % sanitized_table_name)
            sql = "CREATE TABLE %s (%s VARCHAR(20) NULL)" % (sanitized_table_name, temp_column)
            self.cursor.execute(sql)
        except Exception as e:
            msg = e.args[0]
            raise OperationalError("Unable to CREATE TABLE %s, %s" % (sanitized_table_name, msg))

        sql = ''
        try:
            for attr in attributes:
                sql = self.constructSql(table, sanitized_table_name, attr)
                self.log.debug("Executing statement: %s" % sql)
                self.cursor.execute(sql)
            sql_drop = "ALTER TABLE %s DROP COLUMN %s" % (sanitized_table_name, temp_column)
            self.cursor.execute(sql_drop)
            self.db.commit()

        except (pyodbc.Error, Exception) as e:
            msg = e.args[1]
            msg2 = "Encountered problem while executing statement: %s" % sql
            raise OperationalError("Unable to create columns for table %s, %s. %s" % (sanitized_table_name, msg, msg2))


    def match_data_types(self, rally_type, column_name=None):
        '''
        '[42000] [Microsoft][ODBC Driver 13 for SQL Server][SQL Server]The data types text and varchar are incompatible in the equal to operator. (402) (SQLExecDirectW)'
        The text datatype is deprecated and should not be used for new development work.
        '''
        return {
              'INTEGER'    : 'bigint'
            , 'OBJECT'     : 'bigint'
            , 'DATE'       : 'datetime'
            , 'BOOLEAN'    : 'bit default ((0))'
            , 'QUANTITY'   : 'numeric'
            , 'DECIMAL'    : 'decimal(8,3)'
            , 'STRING'     : 'nvarchar(256)'
            , 'TEXT'       : 'nvarchar(max)'
            , 'COLLECTION' : 'varchar(max)'
        }[rally_type]

    def convert_list_to_string_of_quoted_items(self, values):
        # Example of a list received as arg to this method:
        # ['Submitted', 'Open', 'Fixed', 'Closed']
        #
        # Example of a string that we need to pass to
        # ADD COLUMN %s text check (%s IN (%s) to populate IN (%s):
        # 'Submitted','Open','Fixed','Closed'

        str = ""
        for i, item in enumerate(values):
            str += "'" + item + "'"
            if i < len(values) - 1:
                str += ','
        return str

    def constructSqlForObjects(self, table_name, sanitized_table_name, attr):
        data_type = 'bigint'
        _name = self.sanitizeColumnName(attr.ElementName)
        _type = attr.AttributeType
        _values = attr.AllowedValues

        self.log.debug("MSSQLDBConn.constructSqlForObjects(%s, %s, %s) attr.AllowedValues: %s" % (table_name, sanitized_table_name, attr.ElementName, repr(attr.AllowedValues)))

        if attr.SchemaType == 'User':
            sql = "ALTER TABLE %s ADD %s %s" % (sanitized_table_name, _name, self.match_data_types(_type))
        elif attr.ElementName == 'State':
            self.log.debug("MSSQLDBConn.constructSqlForObjects(%s) inside State condition" % (attr.ElementName))
            sql = "ALTER TABLE %s ADD %s nvarchar(256) check (%s IN (%s)) " \
                  % (sanitized_table_name, _name, _name, self.convert_list_to_string_of_quoted_items(_values))
        else:
            sql = "ALTER TABLE %s ADD %s %s" % (sanitized_table_name, _name, data_type)
        return sql

    def constructSql(self, table_name, sanitized_table_name, attr):
        _name   = self.sanitizeColumnName(attr.ElementName)
        _type   = attr.AttributeType
        _values = attr.AllowedValues
        sql = ''

        if _type == 'RATING':
            rating_allowed_values = [a.StringValue for a in _values]
            sql = "ALTER TABLE %s ADD %s nvarchar(256) check (%s IN (%s)) " \
                  % (sanitized_table_name, _name, _name, self.convert_list_to_string_of_quoted_items(rating_allowed_values))
        elif _type == 'STATE':
            state_allowed_values = [a.StringValue for a in _values]
            sql = "ALTER TABLE %s ADD %s nvarchar(256) check (%s IN (%s)) " \
                  % (sanitized_table_name, _name, _name, self.convert_list_to_string_of_quoted_items(state_allowed_values))
        elif _type == 'OBJECT':
            sql = self.constructSqlForObjects(table_name, sanitized_table_name, attr)
        elif _name == 'FormattedID':
            sql = "ALTER TABLE %s ADD %s %s" % (sanitized_table_name, _name, 'varchar(10)')
        elif _name == 'UserPermissions':
            sql = "ALTER TABLE %s ADD %s %s" % (sanitized_table_name, _name, self.match_data_types(_type, _name))
        else:
            sql = "ALTER TABLE %s ADD %s %s" % (sanitized_table_name, _name, self.match_data_types(_type))
        return sql


    def populateFromCSV(self, table_name, primary_key_field, csv_file):
        sanitized_table_name = self.sanitizeTableName(table_name)
        sanitized_table_name = "%s_%s" % (sanitized_table_name, self.table_stamp) if self.table_stamp else sanitized_table_name
        if not os.path.exists(csv_file):
            raise OperationalError("CSV filepath: %s does not exist." % csv_file)
        csv_file_full_path = os.path.join(os.getcwd(), csv_file)
        self.log.debug("Copying data from: %s to table: %s" % (csv_file_full_path, sanitized_table_name))
        # CODEPAGE='65001' worked with french, german, chinese and japanese characters
        sql = """BULK INSERT %s FROM '%s' WITH (FIELDTERMINATOR='\t', ROWTERMINATOR='\n', CODEPAGE='65001'); """ % \
              (sanitized_table_name, csv_file)
        try:
            self.log.debug("Executing statement: %s" % sql)
            self.cursor.execute(sql)
        except (pyodbc.Error, Exception) as e:
            msg = e.args[1]
            print("Problem in copying... %s" % msg)
            raise OperationalError("Unable to populate table %s, %s" % (sanitized_table_name, msg))
        sql1 = "ALTER TABLE %s ALTER COLUMN %s bigint NOT NULL" % (sanitized_table_name, primary_key_field)
        sql2 = "ALTER TABLE %s ADD PRIMARY KEY (%s);" % (sanitized_table_name, primary_key_field)
        self.log.debug("Executing statement: %s" % sql2)
        try:
            self.cursor.execute(sql1)
            self.cursor.execute(sql2)
            self.db.commit()
        except (pyodbc.Error, Exception) as e:
            msg = e.args[1]
            raise OperationalError("Unable to ADD PRIMARY KEY for table %s, %s." % (sanitized_table_name, msg))
