import os,sys
from datetime import datetime
import subprocess
import psycopg2

from dbeif.connection import Connection
from dbeif.utils.eif_exception import OperationalError, ConfigurationError

__version__ = '1.1.0'

class PostgresDBConnection(Connection):

    def __init__(self, config, logger, metrics_recorder=None):
        self.config = config
        self.mr     = metrics_recorder
        super().__init__(logger)
        self.backend_version = None
        self.internalizeConfig(config)

    def name(self):
        return 'dbeif.DBConnection'

    def version(self):
        return __version__


    def internalizeConfig(self, config):
        super().internalizeConfig(config)

        self.server       = config.get('Server', False)
        self.port         = config.get('Port',   5432)
        self.type         = config.get('Type',   None)
        #self.user         = config.get('User',   None)
        #self.password     = config.get('Password', None)
        self.db_name      = config.get('Name',   None)
        self.populate     = config.get('Populate', True)
        self.save_csv     = config.get('SaveCSV',  False)
        self.drop_tables  = config.get('DropExistingTables', False)
        self.stamp_tables = config.get('TimeStampTables', False)
        self.debug        = config.get("Debug",  False)

        # self.proxy = None
        # if self.proxy_server:
        #     self.proxy  = "%s://%s:%s" % (self.proxy_protocol, self.proxy_server, self.proxy_port)
        #     if self.proxy_username and self.proxy_password:
        #         self.proxy  = "%s://%s:%s@%s:%s" % (self.proxy_protocol, self.proxy_username, self.proxy_password, self.proxy_server, self.proxy_port)

        valid_config_items = ['Type','Name','Server', 'Port', 'User', 'Password','Populate',
                              'DropExistingTables','TimeStampTables','SaveCSV','Debug']
        invalid_config_items = [item for item in config.keys() if item not in valid_config_items]
        if invalid_config_items:
            problem = "DBConnection section of the config contained these invalid entries: %s" % ", ".join(invalid_config_items)
            raise ConfigurationError(problem)

    def connect(self):
        try:
            db = psycopg2.connect(host=self.server, port=self.port, database=self.db_name,
                                  user=self.username, password=self.password)
            self.db = db
            self.cursor = self.db.cursor()
        except (psycopg2.OperationalError, Exception) as ex:
            pg_error = ex.args[0].replace('FATAL: ', '').replace("\n", '')
            problem = "Unable to connect to Postgres at %s:%s  : %s" % (self.server, self.port, pg_error)
            raise ConfigurationError(problem)

    def disconnect(self):
        self.cursor.close()
        self.db.close()

    def getBackendVersion(self):
        self.cursor.execute('SELECT version()')
        version_tuple = self.cursor.fetchone()
        self.backend_version = version_tuple[0].split(' on ')[0]
        return self.backend_version

    def validate(self, entities):

        timestamp = datetime.now().isoformat().replace(':', '-').replace('-', '_').split('.')[0]
        self.table_stamp = timestamp if self.stamp_tables else None

        # are there tables present in the db that match the set to be processed?
        # if yes, is DropExistingTables False? then fail validate

        try:
            self.cursor.execute("SELECT table_name FROM information_schema.tables WHERE table_schema = 'public'")
            table_names = [table[0] for table in self.cursor.fetchall()]
        except psycopg2.Error as e:
            print(e)
            raise ConfigurationError("Unable to successfully query your Postgres DB for table names in the %s database" % self.db_name)

        self.existing_tables = [self.sanitizeTableName(entity) for entity in entities
                                if self.sanitizeTableName(entity).lower() in table_names]

        if self.table_stamp:
            return True

        if self.existing_tables:
            if self.drop_tables:
                self.log.warning('You have tables existing in the DB that will be dropped: %s' % ", ".join(self.existing_tables))
            else:
                self.log.error("Tables already exist, cannot proceed.  These tables need to be dropped or renamed: %s" % (", ".join(self.existing_tables)))
                self.log.error("Refer to the documentation about the TimeStampTables and DropExistingTables configuration optional settings.")
                return False

        return True


    def prepareDatabase(self, table_name):
        # if DropExistingTables True, drop the tables for the configured set of entities
        # use the self.existing_tables
        sanitized_table_name = self.sanitizeTableName(table_name)
        sanitized_table_name = "%s_%s" % (sanitized_table_name, self.table_stamp) if self.table_stamp else sanitized_table_name
        try:
            if self.drop_tables and sanitized_table_name in self.existing_tables:
                self.log.debug("Dropping table: %s" % sanitized_table_name)
                sql = 'DROP TABLE %s cascade' %(sanitized_table_name)
                self.cursor.execute(sql)
        except Exception as e:
            msg = e.args[0]
            raise OperationalError("Unable to DROP TABLE %s, %s" % (sanitized_table_name, msg))

    # User is reserved word in PostgreSQL
    def sanitizeTableName(self, table_name):
         return 'Users' if table_name == 'User' else table_name.replace(" ", "")

    def sanitizeColumnName(self, column_name):
         return 'UserID' if column_name == 'User' else column_name

    def createTableAndColumns(self, table, attributes):
        sanitized_table_name = self.sanitizeTableName(table)
        sanitized_table_name = "%s_%s" % (sanitized_table_name, self.table_stamp) if self.table_stamp else sanitized_table_name

        try:
            self.log.debug("Creating table: %s" % sanitized_table_name)
            self.cursor.execute("CREATE TABLE %s ();" % sanitized_table_name)
        except Exception as e:
            msg = e.args[0]
            raise OperationalError("Unable to CREATE TABLE %s, %s" % (sanitized_table_name, msg))

        sql = ''
        try:
            for attr in attributes:
                sql = self.constructSql(table, sanitized_table_name, attr)
                self.log.debug("Executing statement: %s" % sql)
                self.cursor.execute(sql)
            self.db.commit()
        except Exception as e:
            msg = e.args[0]
            msg2 = "Encountered problem while executing statement: %s" % sql
            raise OperationalError("Unable to create columns for table %s, %s. %s" % (sanitized_table_name, msg, msg2))


    def match_data_types(self, rally_type, column_name=None):
        li = 'bigint []'
        if column_name == 'UserPermissions':
           li = 'text []'
        return {
              'INTEGER'    : 'bigint'
            , 'OBJECT'     : 'bigint'
            , 'DATE'       : 'timestamp with time zone'
            , 'BOOLEAN'    : 'boolean default false'
            , 'QUANTITY'   : 'numeric'
            , 'DECIMAL'    : 'numeric'
            , 'STRING'     : 'text'
            , 'TEXT'       : 'text'
            , 'COLLECTION' : li
        }[rally_type]

    def convert_list_to_string_of_quoted_items(self, values):
        # Example of a list received as arg to this method:
        # ['Submitted', 'Open', 'Fixed', 'Closed']
        #
        # Example of a string that we need to pass to
        # ADD COLUMN %s text check (%s IN (%s) to populate IN (%s):
        # 'Submitted','Open','Fixed','Closed'

        str = ""
        for i, item in enumerate(values):
            str += "'" + item + "'"
            if i < len(values) - 1:
                str += ','
        return str

    def constructSqlForObjects(self, table_name, sanitized_table_name, attr):
        data_type = 'bigint'
        _name = self.sanitizeColumnName(attr.ElementName)
        _type = attr.AttributeType
        _values = attr.AllowedValues

        self.log.debug("PGDBConn.constructSqlForObjects(%s, %s, %s) attr.AllowedValues: %s" % (table_name, sanitized_table_name, attr.ElementName, repr(attr.AllowedValues)))

        if attr.SchemaType == 'User':
            sql = "ALTER TABLE %s ADD COLUMN %s %s" % (sanitized_table_name, _name, self.match_data_types(_type))
        elif attr.ElementName == 'State':
            self.log.debug("PGDBConn.constructSqlForObjects(%s) inside State condition" % (attr.ElementName))
            sql = "ALTER TABLE %s ADD COLUMN %s text check (%s IN (%s)) " \
                  % (sanitized_table_name, _name, _name, self.convert_list_to_string_of_quoted_items(_values))
        else:
            sql = "ALTER TABLE %s ADD COLUMN %s %s" % (sanitized_table_name, _name, data_type)
        return sql

    def constructSql(self, table_name, sanitized_table_name, attr):
        _name   = self.sanitizeColumnName(attr.ElementName)
        _type   = attr.AttributeType
        _values = attr.AllowedValues
        sql = ''

        if _type == 'RATING':
            rating_allowed_values = [a.StringValue for a in _values]
            sql = "ALTER TABLE %s ADD COLUMN %s text check (%s IN (%s)) " \
                  % (sanitized_table_name, _name, _name, self.convert_list_to_string_of_quoted_items(rating_allowed_values))
        elif _type == 'STATE':
            state_allowed_values = [a.StringValue for a in _values]
            sql = "ALTER TABLE %s ADD COLUMN %s text check (%s IN (%s)) " \
                  % (sanitized_table_name, _name, _name, self.convert_list_to_string_of_quoted_items(state_allowed_values))
        elif _type == 'OBJECT':
            sql = self.constructSqlForObjects(table_name, sanitized_table_name, attr)
        elif _name == 'FormattedID':
            sql = "ALTER TABLE %s ADD COLUMN %s %s" % (sanitized_table_name, _name, 'text')
        elif _name == 'UserPermissions':
            sql = "ALTER TABLE %s ADD COLUMN %s %s" % (sanitized_table_name, _name, self.match_data_types(_type, _name))
        else:
            sql = "ALTER TABLE %s ADD COLUMN %s %s" % (sanitized_table_name, _name, self.match_data_types(_type))
        return sql


    def populateFromCSV(self, table_name, primary_key_field, csv_file):
        sanitized_table_name = self.sanitizeTableName(table_name)
        sanitized_table_name = "%s_%s" % (sanitized_table_name, self.table_stamp) if self.table_stamp else sanitized_table_name
        if not os.path.exists(csv_file):
            raise OperationalError("CSV filepath: %s does not exist." % csv_file)
        csv_file_full_path = os.path.join(os.getcwd(), csv_file)
        self.log.debug("Copying data from: %s to table: %s" % (csv_file_full_path, sanitized_table_name))
        try:
            csvf = open(csv_file_full_path, 'r', encoding='utf-8')
            copy_command = "COPY %s FROM STDIN NULL '' DELIMITER ',' CSV;" % sanitized_table_name
            self.cursor.copy_expert(copy_command, csvf)
            csvf.close()
        except psycopg2.Error as e:
            print("psycopg2 Problem in copying... %s" % e)
            raise OperationalError("Unable to populate table %s, %s" % (sanitized_table_name, e))
        sql1 = "ALTER TABLE %s ALTER COLUMN %s SET NOT NULL" % (sanitized_table_name, primary_key_field)
        sql2 = "ALTER TABLE %s ADD PRIMARY KEY (%s);" % (sanitized_table_name, primary_key_field)
        self.log.debug("Executing statement: %s" % sql2)
        try:
            self.cursor.execute(sql1)
            self.cursor.execute(sql2)
            self.db.commit()
        except Exception as e:
            msg = e.args[0]
            raise OperationalError("Unable to ADD PRIMARY KEY for table %s, %s." % (sanitized_table_name, msg))
